/*
 * classify_msoffice - classify an ms-office document
 *
 * Copyright (c) Christopher M. R. Lowth - chris@lowth.com
 * Written by Christopher M. R. Lowth
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. The author is not responsible for the consequences of use of this
 *    software, no matter how awful, even if they arise from flaws in it.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Since few users ever read sources,
 *    credits must appear in the documentation.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.  Since few users
 *    ever read sources, credits must appear in the documentation.
 *
 * 4. This notice may not be removed or altered.
 */

/****************************************************************************
 * This program inspects (reads) a specified file, and searches if for
 * predefined sets of strings and binary data - in order to identify what
 * kind of document it is. The "discovered" type is output on stdout as a
 * mime-type string.
 */


#include <fcntl.h>
#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

char 	k1[] = "R\0o\0o\0t\0 \0E\0n\0t\0r\0y",
	k2[] = "O\0b\0j\0e\0c\0t\0P\0o\0o\0l",
	k3[] = "W\0o\0r\0d\0D\0o\0c\0u\0m\0e\0n\0t",
	k4[] = "W\0o\0r\0k\0b\0o\0o\0k",
	k5[] = "V\0B\0A\0_\0P\0R\0O\0J\0E\0C\0T\0_\0C\0U\0R",
	k6[] = "N\0u\0m\0b\0e\0r\0i\0n\0g",
	k7[] = "T\0h\0i\0s\0W\0o\0r\0k\0b\0o\0o\0k",
	k8[] = "P\0R\0O\0J\0E\0C\0T",
	k9[] = "P\0o\0w\0e\0r\0P\0o\0i\0n\0t\0 \0D\0o\0c\0u\0m\0e\0n\0t\0";

struct {
	char *key;				/* Key text */
	int keylen;				/* 0 = Null terminated string */
	int count;				/* Number found */
} keys[] = {
	{ "Excel.Sheet.8" },			/* 0 */
	{ "Microsoft Excel Worksheet" },	/* 1 */
	{ "MSWordDoc" },			/* 2 */
	{ "Word.Document.6" },			/* 3 */
	{ "Word.Document.8" },			/* 4 */
	{ k1, sizeof(k1) },			/* 5  'Root Entry' */
	{ k2, sizeof(k2) },			/* 6  'ObjectPool' */
	{ k3, sizeof(k3) },			/* 7  'WordDocument' */
	{ k4, sizeof(k4) },			/* 8  'Workbook' */
	{ k5, sizeof(k5) },			/* 9  'VBA_PROJECT_CUR' */
	{ k6, sizeof(k6) },			/* 10 'Numbering' */
	{ k7, sizeof(k7) },			/* 11 'ThisWorkbook' */
	{ k8, sizeof(k8) },			/* 12 'PROJECT' */
	{ "ThisWorkbook" },			/* 13 */
	{ k9, sizeof(k9) },			/* 14 */
	{ "Microsoft PowerPoint" },
	{ "Slide Titles" },
	{ NULL }
};


struct {
	char *key;
	char *type;
} types[] = {
	{ "00000100000000000",	"application/vnd.ms-excel 00000100000000000" },
	{ "00000100000011000",	"application/vnd.ms-excel 00000100000011000" },
	{ "00000100100000000",	"application/vnd.ms-excel 00000100100000000" },
	{ "00000111000000000",	"application/msword 00000111000000000" },
	{ "00001111000000000",	"application/msword 00001111000000000" },
	{ "00101101000000000",	"application/msword Version 6" },
	{ "00101101000010000",	"application/msword 00101101000010000" },
	{ "00101111000000000",	"application/msword 00101111000000000" },
	{ "00101111000010000",	"application/msword 00101111000010000" },
	{ "00110101000000000",	"application/msword 00110101000000000" },
	{ "00110111000000000",	"application/msword 00110111000000000" },
	{ "00111101000000000",	"application/msword 00111101000000000" },
	{ "01000100000011000",	"application/vnd.ms-excel 01000100000011000" },
	{ "10101111000000000",	"application/msword 10101111000000000" },
	{ "11000100110111000",	"application/vnd.ms-excel 11000100110111000" },
	{ "11000100111111000",	"application/vnd.ms-excel 11000100111111000" },
	{ "00000100000000111",	"application/vnd.ms-powerpoint 00000100000000111" },

	{ NULL }
};

int findkey(char *buff, int len, char *key, int keylen)
{
	int i;
	int n = 0;

	if (keylen == 0)
		keylen = strlen(key);

	for (i=0; i<(len - keylen); i++)
		if (memcmp(buff + i, key, keylen) == 0)
			n ++;
	return n;
}

int main(int argc, char *argv[])
{
	int id;
	int len;
	int i;
	static unsigned char magic[] = {
		0xD0, 0xCF, 0x11, 0xE0, 0xA1, 0xB1, 0x1A, 0xE1
	};
	char buff[40960];
	char str[80];

	for (i=0; keys[i].key; i++) {
		if (keys[i].keylen == 0)
			keys[i].keylen = strlen(keys[i].key);
	}

	id = open(argv[1], O_RDONLY);
	if (id == -1)
		return 2;
	len = read(id, buff, 8);
	assert(len == 8);
	if (memcmp(buff, magic, 8) != 0)
		return 3;	/* Not an MS-office document */
	for ( ; ; ) {
		len = read(id, buff, sizeof(buff));

		for (i=0; keys[i].key != NULL; i++)
			keys[i].count += findkey(buff, len, keys[i].key,
							keys[i].keylen);

		if (len < sizeof(buff))
			break;
		lseek(id, -64, SEEK_CUR);	
	}
	close(id);

	str[0] = '\0';
	for (i=0; keys[i].key; i++) {
		if (keys[i].count > 0)
			strcat(str, "1");
		else
			strcat(str, "0");
	}

	for (i=0; types[i].key; i++) {
		if (strcmp(str, types[i].key) == 0) {
			printf("%s\n", types[i].type);
			return 0;
		}
	}

	printf("application/x-unknown-msoffice %s\n", str);
	return 1;
}
