/*
 * Copyright (c) Christopher M. R. Lowth - chris@lowth.com
 * Written by Christopher M. R. Lowth
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. The author is not responsible for the consequences of use of this
 *    software, no matter how awful, even if they arise from flaws in it.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Since few users ever read sources,
 *    credits must appear in the documentation.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.  Since few users
 *    ever read sources, credits must appear in the documentation.
 *
 * 4. This notice may not be removed or altered.
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <assert.h>

#include "protector.h"

/****************************************************************************/

static int add_header(mailrec_t *mailrec, char *text)
{
	field_t **fp;
	char *p;
	char *key = NULL;
	char *val = NULL;
	field_t *fld = NULL;

	assert(text != NULL);
	assert(mailrec != NULL);

	/*
	 * Strip leading white space from the text
	 */

	while (*text == ' ' || *text == '\t')
		text ++;

	/*
	 * Take malloced copy of the text
	 */

	text = strdup(text);
	if (text == NULL) {
		local_errno = E_OUT_OF_MEMORY;
fprintf(stderr, "headers.c: %d: strlen(text)=%d\n", __LINE__, strlen(text));
		return FALSE;
	}

	/*
	 * Search for the current end of the headers chain
	 */

	for (fp = &(mailrec->headers); *fp != NULL; fp = &((*fp)->next))
		;

	if (strncasecmp(text, "From ", 5) == 0) {
		key = strdup(":From:");
		val = strdup(text);
	}
	else {
		p = strchr(text, ':');
		if (p == NULL) {
			local_errno = E_NO_COLON;	
fprintf(stderr, "headers.c: %d: no colon in '%s'\n", __LINE__, text);
			goto failed;
		}
		*p = '\0';

		if (strcasecmp(text, "content-length") == 0) {
			free(text);
			return TRUE;
		}

		p ++;
		while (*p == ' ' || *p == '\t')
			p ++;

		key = strdup(text);
		val = strdup(p);
	}

	if (
		key == NULL || val == NULL ||
		(fld = malloc(sizeof(field_t))) == NULL
	) {
		local_errno = E_OUT_OF_MEMORY;
fprintf(stderr, "headers.c: %d: out of memory\n", __LINE__);
		goto failed;
	}

	fld->key = key;
	fld->value = val;
	fld->next = NULL;

	(*fp) = fld;

	free(text);

	return TRUE;

failed:
	if (key != NULL) free(key);
	if (val != NULL) free(val);
	if (fld != NULL) free(fld);
	if (text != NULL) free(text);
	return FALSE;
}

/****************************************************************************/

static int extend_last_header(mailrec_t *mailrec, char *text)
{
	field_t **fp;
	char *p;
	int len;

	assert(text != NULL);
	assert(mailrec != NULL);

	text = strdup(text);
	if (text == NULL) {
		local_errno = E_OUT_OF_MEMORY;
fprintf(stderr, "headers.c: %d: strlen(text)=%d\n", __LINE__, strlen(text));
		return FALSE;
	}

	if (mailrec->headers == NULL) {
		local_errno = E_BAD_HEADER;
fprintf(stderr, "headers.c: %d: mailrec->headers == NULL\n", __LINE__);
		return FALSE;
	}

	for (fp = &(mailrec->headers); (*fp)->next != NULL; fp = &((*fp)->next))
		;

	len = strlen((*fp)->value) + strlen(text);
	p = malloc(len + 2);
	if (p == NULL) {
		local_errno = E_OUT_OF_MEMORY;
		free(text);
fprintf(stderr, "headers.c: %d: len+2=%d\n", __LINE__, len+2);
		return FALSE;
	}
	strcpy(p, (*fp)->value);
	strcat(p, "\n");
	strcat(p, text);
	free((*fp)->value);
	(*fp)->value = p;

	return TRUE;
}



int read_headers(FILE *source, mailrec_t *mailrec)
{
	char buff[4096];
	char *p;
	int rtn;
	long pos = ftell(source);
	
	/*
	 * Read the mail headers
	 */

	while (fgets(buff, sizeof(buff), source) != NULL) {
		if ((p = strchr(buff, '\r')) != NULL) *p = '\0';
		if ((p = strchr(buff, '\n')) != NULL) *p = '\0';

		if (buff[0] == '\0')
			break;

		if (mailrec->headers == NULL && !isalpha(buff[0])) {
			fseek(source, pos, 0);
			return TRUE;
		}

		if (buff[0] == ' ' || buff[0] == '\t')
			rtn = extend_last_header(mailrec, buff);
		else
			rtn = add_header(mailrec, buff);

		if (!rtn)
			return FALSE;

		pos = ftell(source);
	}

	return TRUE;
}

char *get_header(mailrec_t *mailrec, char *key, int instance)
{
	field_t *f;

	if (instance == -1) {	/* Insist on it being unique */
		char *i1 = get_header(mailrec, key, 1);
		char *i2 = get_header(mailrec, key, 2);
		return i2 == NULL ? i1 : NULL;
	}

	for (f = mailrec->headers; f != NULL; f = f->next) {
		if (strcasecmp(f->key, key) == 0) {
			instance --;
			if (instance == 0)
				return f->value;
		}
	}

	return NULL;
}


char *unquote(char *str)
{
	int slen;
	int open_quote = 0;
	int close_quote = 0;

	str = strdup(str);
	assert(str != NULL);

	/* trim leading white space */	
	while (isspace(*str))
		str ++;

	slen = strlen(str);

	/* trim trailing white space */	
	while (slen > 0 && isspace(str[slen-1]))
		slen --;

	/* dequote */
	if (slen > 0 && (str[0] == '"' || str[0] == '\'')) {
		open_quote = str[0];
		str ++;
		slen --;
	}

	if (slen > 0 && (str[slen-1] == '"' || str[slen-1] == '\'')) {
		close_quote = str[slen-1];
		slen --;
	}

	str[slen] = '\0';

	/* Trap for mal formed headers (mismatched quotes) */

	assert(open_quote == close_quote);	
	if (open_quote) assert(strchr(str, open_quote) == NULL);

	return str;
}

char *get_sub_header(char *text, char *key)
{
	int i, quoted;
	int klen;
	char *dup, *subfield;

	dup = text = strdup(text);

top:
	while (*text && (isspace(*text) || *text == ';'))
		text++;

	if (*text == '\0') {
		free(dup);
		local_errno = E_EMPTY_STRING;
		return NULL;

	}

	/*
	 * Find the length (i) of the key/value combination,
	 * taking the semicolon delimiter into account
	 */
	
	quoted = FALSE;
	for (i=0; text[i]; i++) {
		if (text[i] == '"') {
			quoted = !quoted;
			continue;
		}
		if (!quoted && text[i] == ';')
			break;
	}
	assert(i > 0);

	/* Take a copy of the key/value pair */

	subfield = malloc(i + 1);
	if (subfield == NULL) {
		local_errno = E_OUT_OF_MEMORY;
		free(dup);
		return NULL;
	}
	memcpy(subfield, text, i);
	subfield[i] = '\0';

	/* if we were given a null key - the entire field is the value */

	if (key == NULL)
		return unquote(subfield);

	/* if the field matches our key - return the value */

	klen = strlen(key);	
	if (
		strlen(subfield) >= klen + 1 &&
		strncasecmp(text, key, klen) == 0 &&
		text[klen] == '='
	)
		return unquote(subfield + klen + 1);

	/* not found it yet - loop back */

	text += i;
	if (*text) {
		text ++;
		goto top;
	}

	free(dup);
	local_errno = E_SUB_HEADER_NOT_FOUND;
	return NULL;
}


/****************************************************************************
 * Find out who sent this thing to us - or rather: who we should tell about
 * things if we dont like his mail to us.
 */

char *get_sender(mailrec_t *mailrec)
{
	/*
	 * The order of these header keys is important - they are
	 * prioritised by "procmail" (see formail.c), and so we do the same
	 * (we dont know any better).
	 */

	static char *keys[] = {
		"resent-reply-to",
		"resent-sender",
		"resent-from",
		"return-receipt-to",
		"errors-to",
		"reply-to",
		"sender",
		"from",
		"return-path",
		"path",
		":from:",
		NULL
	};
	int i;
	char *p = NULL;

	for (i=0; keys[i] != NULL; i++) {
		p = get_header(mailrec, keys[i], 1);
		if (p != NULL)
			break;
	}

	if (p == NULL)
		return NULL;

	return p;
}

char *get_recipient(mailrec_t *mailrec)
{
	return get_header(mailrec, "To", 1);
}
