/*
 * read - collection of functions used in reading and parsing an internet
 *        email message. Part of the "protector" program.
 *
 * Copyright (c) Christopher M. R. Lowth - chris@lowth.com
 * Written by Christopher M. R. Lowth
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. The author is not responsible for the consequences of use of this
 *    software, no matter how awful, even if they arise from flaws in it.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Since few users ever read sources,
 *    credits must appear in the documentation.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.  Since few users
 *    ever read sources, credits must appear in the documentation.
 *
 * 4. This notice may not be removed or altered.
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <fcntl.h>
#include <unistd.h>

#include "protector.h"
#include "proto.h"

/****************************************************************************/

static char *get_boundary(mailrec_t *mailrec)
{
	char *content_type = get_header(mailrec, "content-type", 1);
	char *boundary = NULL;

	if (
		content_type == NULL ||
		strncasecmp(content_type, "multipart/", 10) != 0
	)
		return NULL;
	boundary = get_sub_header(content_type, "boundary");
	return boundary;
}

/****************************************************************************/

static void append_part(mailrec_t *mailrec, mailrec_t *part)
{
	mailrec_t **mp;

	for (mp = &(mailrec->parts); *mp != NULL; mp = &((*mp)->next))
		;
	(*mp) = part;
}

/****************************************************************************/

mailrec_t *read_mailrec(FILE *source, char *end_boundary, int stop_at_From)
{
	char buff[4096];
	char *p;
	mailrec_t *mailrec = NULL;
	int boundary_len = 0;
	mailrec_t *part;
	long pos;
	int end_flag;

	if ((mailrec = malloc(sizeof(mailrec_t))) == NULL) {
		local_errno = E_OUT_OF_MEMORY;
		return NULL;
	}

	memset(mailrec, 0, sizeof(mailrec_t));

	mailrec->file = source;
	mailrec->start_header = ftell(source);

	if (!read_headers(source, mailrec)) {
		free_mailrec(mailrec);
		return NULL;
	}

	mailrec->start_body = pos = ftell(source);

	/*
	 * Determine the boundary string to use (if any)
	 */

	mailrec->boundary = get_boundary(mailrec);
	boundary_len = (mailrec->boundary == NULL) ? 0 : strlen(mailrec->boundary);

	/*
	 * Now read the mail body, hunting for boundaries (etc) as we go.
	 * There are 3 circumstances under which we will stop reading..
	 *
	 *	1	we hit end of file
	 *
	 *	2	we hit a "From" line, marking the start of the next
	 *		message (if stop_at_from is true).
	 *
	 *	3	we hit the specified end_boundary (if specified).
	 */

	while(fgets(buff, sizeof(buff), source) != NULL) {
		if ((p = strchr(buff, '\r')) != NULL) *p = '\0';
		if ((p = strchr(buff, '\n')) != NULL) *p = '\0';

		/*
		 * Determine whether we have hit the end of the section we
		 * are scanning, and if so - jump out of the loop.
		 */

		if (
			(
				stop_at_From &&
				strncasecmp(buff, "from ", 5) == 0
			) || (
				end_boundary != NULL &&
				memcmp(buff, "--", 2) == 0 &&
				memcmp(
					buff+2,
					end_boundary,
					strlen(end_boundary)
				) == 0
			)
		)
			break;

		/*
		 * If we have found one of our part boundaries, then make
		 * internal notes so that we can find it again quickly.
		 */

		if (
			mailrec->boundary != NULL &&
			memcmp(buff, "--", 2) == 0 &&
			memcmp(buff+2, mailrec->boundary, boundary_len) == 0
		) {
			if (mailrec->parts == NULL)
				mailrec->start_parts = pos;

			end_flag = strcmp(buff + 2 + boundary_len, "--") == 0;

			part = read_mailrec(
				source,
				end_flag ? "" : mailrec->boundary,
				stop_at_From
			);
			append_part(mailrec, part);
			part->last_part = end_flag;
		}

		pos = ftell(source);
	}

	mailrec->end = pos;

	if (mailrec->parts == NULL)
		mailrec->start_parts = mailrec->end;

	fseek(source, pos, 0);

	return mailrec;
}
