/*
 * write - collection of functions used in writing out (or filtering) an
 *         internet e-mail message. Part of the "protector" program.
 *
 * Copyright (c) Christopher M. R. Lowth - chris@lowth.com
 * Written by Christopher M. R. Lowth
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. The author is not responsible for the consequences of use of this
 *    software, no matter how awful, even if they arise from flaws in it.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Since few users ever read sources,
 *    credits must appear in the documentation.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.  Since few users
 *    ever read sources, credits must appear in the documentation.
 *
 * 4. This notice may not be removed or altered.
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

#include "protector.h"
#include "proto.h"

void dump_mailrec(mailrec_t *mailrec, int padlen)
{
	field_t *h;
	mailrec_t *p;
	char pad[80];

	memset(pad, ' ', sizeof(pad));
	pad[padlen * 4] = '\0';

	printf("%s{\n", pad);
	printf("%sfile         = %p\n", pad, mailrec->file);
	printf("%sstart_header = %ld\n", pad, mailrec->start_header);
	printf("%sstart_body   = %ld\n", pad, mailrec->start_body);
	printf("%sstart_parts  = %ld\n", pad, mailrec->start_parts);
	printf("%send          = %ld\n", pad, mailrec->end);
	printf("%sheaders      = %p\n", pad, mailrec->headers);
	printf("%sboundary     = %s\n", pad, mailrec->boundary);
	printf("%sparts        = %p\n", pad, mailrec->parts);
	printf("%snext         = %p\n", pad, mailrec->next);
	printf("%s---- headers ----\n", pad);
	for (h = mailrec->headers; h != NULL; h = h->next) {
		printf("%s%s:%s\n\n", pad, h->key, h->value);
	}
	printf("%s---- parts ----\n", pad);
	for (p = mailrec->parts; p != NULL; p = p->next)
		dump_mailrec(p, padlen + 1);
	printf("%s}\n", pad);
}

void put_headers_in_environment(mailrec_t *mailrec)
{
	field_t *h;
	char buff[8192];
	int i;
	char *H, *hdr, *fld;
	int len;

	len = 0;	
	for (h = mailrec->headers; h != NULL; h = h->next) {
		strcpy(buff, "H_");
		strcat(buff, h->key);
		for (i=0; buff[i]; i++) {
			buff[i] = toupper(buff[i]);
			if (!isalnum(buff[i]))
				buff[i] = '_';
		}
		strcat(buff, "=");
		strcat(buff, h->value);
		for (i=0; buff[i]; i++) {
			if (isspace(buff[i]))
				buff[i] = ' ';
		}
		putenv(strdup(buff));
		len += strlen(h->key) + strlen(h->value) + 8;
	}

	H = malloc(len + 8);
	assert(H != NULL);
	memset(H, 0, len+8);

	strcat(H, "HEADERS=");

	for (h = mailrec->headers; h != NULL; h = h->next) {
		strcat(H, h->key);
		strcat(H, ": ");
		strcat(H, h->value);
		strcat(H, "\n");
	}

	putenv(H);

	putenv("LIB_DIR=" LIB_DIR);


	if ((hdr = get_header(mailrec, "content-type", 1)) == NULL) hdr = "";
	if ((fld = get_sub_header(hdr, NULL)) == NULL) fld = "";
	fld = strdup(fld);
	for (i=0; fld[i]; i++) fld[i] = tolower(fld[i]);
	sprintf(buff, "CONTENT_TYPE=%s", fld);
	putenv(strdup(buff));


	if ((fld = get_sub_header(hdr, "name")) == NULL) fld = "";
	sprintf(buff, "CONTENT_TYPE__NAME=%s", fld);
	putenv(strdup(buff));


	if ((hdr = get_header(mailrec, "content-disposition", 1)) == NULL)
		hdr = "";
	if ((fld = get_sub_header(hdr, "filename")) == NULL) fld = "";
	sprintf(buff, "CONTENT_DISPOSITION__FILENAME=%s", fld);
	putenv(strdup(buff));

	if ((fld = get_header(mailrec, "content-transfer-encoding", 1)) == NULL)
		fld = "";
	fld = strdup(fld);
	for (i=0; fld[i]; i++) fld[i] = tolower(fld[i]);
	sprintf(buff, "CONTENT_TRANSFER_ENCODING=%s", fld);
	putenv(strdup(buff));

}

/*
	99	cant execl() the filter script.
	98	filter script failed oddly (crashed?).
	97	waitpid on script gave wrong return.

	Others	as returned by the script itself.
*/

int filter_mailrec(FILE *target, mailrec_t *mailrec)
{
	int p1[2], p2[2];
	FILE *f;
	pid_t pid1, pid2;
	char *prog = LIB_DIR "/bin/part_filter";
	char *ct;
	int got;
	char buff[8192];
	int rtn, status, exit_code = 0;

	ct = get_header(mailrec, "content-type", 1);
	if (ct != NULL && strncasecmp(ct, "multipart/", 10) == 0)
		return write_mailrec(target, mailrec, TRUE);

	pipe(p1);	/* p1[0] = filter's stdin */
	pipe(p2);	/* p2[1] = filter's stdout */

	/*
	 * The filter process
	 */

	pid1 = fork();
	if (pid1 == 0) {
		close(p1[1]); close(p2[0]);
		close(0); dup(p1[0]); close(p1[0]);
		close(1); dup(p2[1]); close(p2[1]);

		put_headers_in_environment(mailrec);
		execl(prog, prog, NULL);

		_exit(99);
	}

	/*
	 * Process to write to filter's stdin (p1)
	 */

	pid2 = fork();
	if (pid2 == 0) {
		close(p1[0]); close(p2[0]); close(p2[1]);
		f = fdopen(p1[1], "w");
		rtn = write_mailrec(f, mailrec, FALSE);
		fclose(f);
		_exit(rtn);
	}

	/*
	 * The parent reads the filter's stdout (p2)
	 */

	close(p1[1]); close(p1[0]); close(p2[1]);

	f = fdopen(p2[0], "r");
	while ((got = fread(buff, 1, sizeof(buff), f)) > 0)
		fwrite(buff, 1, got, target);
	fclose(f);

	if (waitpid(pid1, &status, 0) == pid1) {
		if (WIFEXITED(status))
			exit_code = WEXITSTATUS(status);
		else
			exit_code = 98;
	}
	else
		exit_code = 97;

	waitpid(pid2, NULL, 0);

	return exit_code;
}


int write_mailrec(FILE *target, mailrec_t *mailrec, int include_headers)
{
	field_t *h;
	mailrec_t *part;
	long len;
	int n, got;
	char buff[8192];
	int rtn;
	int rtn_code = 0;

	/*
	 * Output the headers (if wanted)
	 */

	if (include_headers) {
		for (h = mailrec->headers; h != NULL; h = h->next) {
			if (h->key[0] != ':')
				fprintf(target, "%s: ", h->key);
			fprintf(target, "%s\n", h->value);
	
		}

		fprintf(target, "\n");
	}

	/*
	 * Output the part of the mail message BEFORE any boundaries
	 */

	fseek(mailrec->file, mailrec->start_body, 0);
	len = mailrec->start_parts - mailrec->start_body;
	while (len > 0) {
		n = (len > sizeof(buff)) ? sizeof(buff) : len;
		if ((got = fread(buff, 1, n, mailrec->file)) < 1)
			break;
		fwrite(buff, 1, got, target);
		len -= got;
	}

	/*
	 * Output the attachments (parts). All but the last "empty" one get
	 * passed to our helper filter program or script for checking
	 * and/or modifying.
	 */

	for (part = mailrec->parts; part != NULL; part = part->next) {
		fprintf(target, /* "\n" */ "--%s%s\n", mailrec->boundary,
			part->last_part ? "--" : "");
		if (part->last_part)
			rtn = write_mailrec(target, part, TRUE);
		else
			rtn = filter_mailrec(target, part);
		if (rtn_code == 0)
			rtn_code = rtn;
	}

	/*
	 * Happy bunny! It all worked just fine.
	 */

	return rtn_code;
}
